//**********************************************************************************************************
//
//   Version 1.0.0
//
// The maxconn exit is an MQSeries security exit that allows client connections to be limited according to 
// the name of the associated ServerConnection channel.  
// Each Server Connection channel can be allocated a maximum number of connections and the maxconn exit will ensure 
// that this maximum is not exceeded. 
//
// Client connections to a queue manager are limited by the MaxChannels property. If the client connections 
// originate from multiple applications, then it is a possibility that one application could "swamp" the queue 
// manager with client connections, thus preventing any connections being made from other applications.
// The maxconn exit can be employed to limit the number of connections available to each application. 
//
// author	Tony Madden	tm@mqware.com
//**********************************************************************************************************
 #include <cmqc.h>
 #include <cmqxc.h>
 #include <cmqbc.h>
#include <cmqcfc.h>

#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdarg.h>
#include	<time.h>


#include	<maxconn.h>
//#include	<conmaxutil.h>

#define	MXC_EXIT_PROG_IDENT	"maxconn v1.0.1" 

typedef struct	workfields{

	PMQCXP		pMQCXP;
	PMQCD		pMQCD;
	PMQCHAR		pcAgentBuffer;

	MQHCONN		hConn;
	MQHOBJ		dataQHobj;
	MQHOBJ		eventQHobj;
	MQHOBJ		waitQHobj;

	MXC_STATEDATA	sd;

	char		szQMgr[MQ_Q_MGR_NAME_LENGTH+1];
	char		szChannelName[MQ_CHANNEL_NAME_LENGTH+1];
	char		szConnectionName[MQ_SHORT_CONN_NAME_LENGTH+1];
	char		szRemoteUserId[MQ_USER_ID_LENGTH+1];
	char		szDataQueueName[MQ_Q_NAME_LENGTH+1];
	char		szVolatileCorrelId[MQ_CORREL_ID_LENGTH+1];
	char		szBootstrapCorrelId[MQ_CORREL_ID_LENGTH+1];
	char		szLogFile[257];
//	char		szLogFileName[257];
//	char		szLogFileDirectory[257];

	char		szUpdatedVolatileData[500];

	long		lOpenType;
	long		lCallType;

} MXC_WORKFIELDS,*pMXC_WORKFIELDS;



 
long	bagSetup(pMXC_WORKFIELDS,MQHBAG*const,MQHBAG*const,MQHBAG*const);
long	closeQ(pMXC_WORKFIELDS);
long	connectToQMgr(pMXC_WORKFIELDS);
long	connectionDecision(pMXC_WORKFIELDS,long,long*,long*);
long	eventDecision(pMXC_WORKFIELDS ,long );
long	extractValues(pMXC_WORKFIELDS,char*,long);
long	decideAndUpdate(pMXC_WORKFIELDS ,long);
long	getActualCurrentNumberOfRunningChannels(pMXC_WORKFIELDS,long*);
long	getBootstrapData(pMXC_WORKFIELDS,char*,long*);
long	getVolatileData(pMXC_WORKFIELDS,char*,long*);
long	initialize(pMXC_WORKFIELDS,PMQCXP,PMQCD,PMQCHAR,long);
long	mqEvent(pMXC_WORKFIELDS,long);
void	nullDelimit(char[],char[],long);
long	openQueue(pMXC_WORKFIELDS,char[],long,MQHOBJ*);
void	processConnectionRequest1(PMQCXP,PMQCD,PMQCHAR);
void	processConnectionRequest2(PMQCXP);
void	processConnectionTermination(PMQCXP,PMQCD,PMQCHAR);
long	reWriteVolatileData(pMXC_WORKFIELDS);
long	sendMqEventMessage(pMXC_WORKFIELDS, char*,long);
long	waitForExclusiveLockToEnd(pMXC_WORKFIELDS,long);
void	writeError(pMXC_WORKFIELDS,char data[],...);
void	writeLog(pMXC_WORKFIELDS,char[],char[],...);
void	writeToFile(pMXC_WORKFIELDS,char[]);
 
void	MQENTRY maxconn(PMQCXP,PMQCD,PMQLONG ,PMQLONG, PMQCHAR,PMQLONG,PMQCHAR*);

void	MQStart() {;}

void MQENTRY maxconn
                    (PMQCXP   arg_pMQCXP
                    ,PMQCD    arg_pMQCD
                    ,PMQLONG  arg_plDataLength
                    ,PMQLONG  arg_plAgentBufferLength
                    ,PMQCHAR  arg_pcAgentBuffer
                    ,PMQLONG  arg_plExitBufferLength
                    ,PMQCHAR *arg_ppcExitBuffer
                    )



{

	 
//	if (arg_pMQCXP->ExitId == MQXT_CHANNEL_SEC_EXIT && arg_pMQCD->ChannelType == MQCHT_SVRCONN){
	if (arg_pMQCXP->ExitReason == MQXR_INIT){
		arg_pMQCXP->ExitResponse = MQXCC_OK;
	}else{
		if (arg_pMQCXP->ExitReason == MQXR_INIT_SEC){
			arg_pMQCXP->ExitResponse = MQXCC_SEND_SEC_MSG;
		}else{
			if (arg_pMQCXP->ExitReason == MQXR_SEC_MSG ){
				//  determine whether to suspend or allow the connection request
				processConnectionRequest1(arg_pMQCXP,arg_pMQCD,arg_pcAgentBuffer);
			}else{
				if(arg_pMQCXP->ExitReason == MQXR_TERM){
					// Perform disconnect processing
					processConnectionTermination(arg_pMQCXP,arg_pMQCD,arg_pcAgentBuffer);
				}
			}
		}
	}

	
//	}
}

//**********************************************************************************************************
//  Determine the existing number of connections using the arg_pMQCD->ChannelName channel
//  Determine whether the current connection request should be allowed or refused
//  Store the decision in arg_pMQCXP->ExitUserArea
//
//**********************************************************************************************************
void processConnectionRequest1(PMQCXP   arg_pMQCXP,
							   PMQCD    arg_pMQCD,
							   PMQCHAR	arg_pcAgentBuffer){
	
	long		lRc = MXC_RC_NONE;
	long		lDataGotFromBootstrapStore = 0;
	long		lBootstrap = 0;
	long		lLength;

	char		buffer[1000];	
	
	MXC_WORKFIELDS	wf;

	// Initialize variables and open the appropriate queue
	lRc = initialize(&wf,arg_pMQCXP,arg_pMQCD,arg_pcAgentBuffer,MXC_REQUEST);

	if (lRc == MXC_RC_NONE){
		// Volatile data contains existing state (e.g. number of existing connections prior to current request) 
		lLength = sizeof(buffer);
		lRc = getVolatileData(&wf,(char*)buffer,&lLength);
		if (lRc == MXC_RC_NONE){	
			// Parse the volatile data
			lRc = extractValues(&wf,(char*)buffer,lLength);
		}
		if (lRc == MXC_RC_NO_DATA_AVAILABLE){
			//  If no volatile data, then retrieve the bootstrap data and use that instead.
			lLength = sizeof(buffer);
			lRc = getBootstrapData(&wf,(char*)buffer,&lLength);
			if (lRc == MXC_RC_NONE){
				lDataGotFromBootstrapStore = 1;
				// Parse the bootstrap data
				lRc = extractValues(&wf,(char*)buffer,lLength);
				lBootstrap = 1;
			}
		}
	}

	if (lRc == MXC_RC_NONE){
		if ((wf.sd.lTraceLevel & MXC_TRACE_LEVEL_DEBUG) == MXC_TRACE_LEVEL_DEBUG){
			buffer[lLength] = 0;
			if (lBootstrap){
				writeLog(&wf,MXC_LOG_DEBUG,(char*)"%s. Connection Process Commencing . Bootstrap Data <%s>. ExtractedValues: MaxConnections=%ld, ExistingConnections=%ld, TraceLevel=%ld,  EventState=%ld ",MXC_EXIT_PROG_IDENT,buffer,wf.sd.lMaxConnections,wf.sd.lExistingConnections,wf.sd.lTraceLevel,wf.sd.lEventState);
			}else{
				writeLog(&wf,MXC_LOG_DEBUG,(char*)"Connection Process Commencing . Volatile Data <%s>. ExtractedValues: MaxConnections=%ld, ExistingConnections=%ld, TraceLevel=%ld,  EventState=%ld ",buffer,wf.sd.lMaxConnections,wf.sd.lExistingConnections,wf.sd.lTraceLevel,wf.sd.lEventState);
			}
		}
		// Allow the connection or not? Update volatile data accordingly.
		decideAndUpdate(&wf,lDataGotFromBootstrapStore);
	}

	closeQ(&wf); 
		
	arg_pMQCXP->ExitUserArea[MXC_EXITUSERAREA_TRACELEVEL_OFFSET] = (int)wf.sd.lTraceLevel; 

	

}

//**********************************************************************************************************
//  This function is called when the exit is called with MQXR_TERM
//  If the arg_pMQCXP->ExitUserArea[0] specifies that the connection is rejected (MXC_REJECTED) , then no action
//  If the arg_pMQCXP->ExitUserArea[0] specifies that the connection is accepted (MXC_ACCEPTED), then retrieve the volatile data and decrement connection count
//**********************************************************************************************************
void processConnectionTermination(PMQCXP	arg_pMQCXP,
								  PMQCD		arg_pMQCD,
								  PMQCHAR	arg_pcAgentBuffer){

	int iTrace = arg_pMQCXP->ExitUserArea[MXC_EXITUSERAREA_TRACELEVEL_OFFSET]; 

	if (arg_pMQCXP->ExitUserArea[MXC_EXITUSERAREA_RESPONSE_OFFSET] == MXC_REJECTED && ((iTrace & MXC_TRACE_LEVEL_INFO ) != MXC_TRACE_LEVEL_INFO)){
		// If the connection was rejected, and trace level does not include debug then there is nothing to do...
		;
	}else{

		long		lRc = MXC_RC_NONE;
		long		lLength;
		char		buffer[1000];	
		MXC_WORKFIELDS	wf;

		lRc = initialize(&wf,arg_pMQCXP,arg_pMQCD,arg_pcAgentBuffer,MXC_TERM);
		if (lRc == MXC_RC_NONE){
			lLength = sizeof(buffer);
			lRc = getVolatileData(&wf,(char*)buffer,&lLength);
			if (lRc == MXC_RC_NONE){	
				lRc = extractValues(&wf,(char*)buffer,lLength);
			}
		}

		if (lRc == MXC_RC_NONE){ 

			if (arg_pMQCXP->ExitUserArea[MXC_EXITUSERAREA_RESPONSE_OFFSET] == MXC_REJECTED){
				// If the connection is rejected then all there is to do is write the log info 
				// Note that trace level must be at least MXC_TRACE_LEVEL_INFO if we get to here, so no need to check trace level...
				writeLog(&wf,MXC_LOG_INFO,(char*)"*** Connection terminated due to refused connection for connection name <%s> using channel %s. Existing=%ld. Max=%ld",wf.szConnectionName,wf.szChannelName,wf.sd.lExistingConnections,wf.sd.lMaxConnections);
			}else{
				long lEvent=0;
				// Decrement the existing channel count by one
				if (--wf.sd.lExistingConnections < 0){
					// How can this ever be? I don't know, but I check just in case....
					wf.sd.lExistingConnections = 0;
				}

				if (wf.sd.lEventFlag){
					// NOTE - eventDecision() must be called before createStateDataString()
					lEvent = eventDecision(&wf,wf.sd.lExistingConnections);
				}
				
				// Construct the message
				createStateDataString(wf.sd,wf.szUpdatedVolatileData,MXC_VOLATILE_PREFIX);
				
				if (lEvent && (wf.sd.lEventFlag & MXC_EVENTS_MQ_ACTIVE) == MXC_EVENTS_MQ_ACTIVE){
					// Create an MQ event message and write it to the event queue
					// NOTE - mqEvent() must be called after createStateDataString() 
					mqEvent(&wf,MXC_CONNECTION_TERMINATION);
				}
				
				// Write volatile data to State queue
				reWriteVolatileData(&wf);

				if ((wf.sd.lTraceLevel & MXC_TRACE_LEVEL_INFO) == MXC_TRACE_LEVEL_INFO){
					writeLog(&wf,MXC_LOG_INFO,(char*)"*** Connection terminated for connection name <%s> using channel %s. Existing=%ld. Max=%ld",wf.szConnectionName,wf.szChannelName,wf.sd.lExistingConnections,wf.sd.lMaxConnections);
				}
		
			}

			closeQ(&wf);
			
		
		}
	}

	arg_pMQCXP->ExitResponse = MQXCC_OK;	// FIX001.  Not really a fix, but this line was moved from the maxconn exported function body.
}

//**********************************************************************************************************
// setup the exit 
//
// arg_pWf				work properties
// arg_pMQCXP			MQCXP
// arg_MQCD				MQCD
// arg_pcAgentBuffer	message data
// arg_lCallType		file specifying whether the exit has been called for a connection or a termination.
//**********************************************************************************************************
long	initialize(MXC_WORKFIELDS*	arg_pWf,
				   PMQCXP		arg_pMQCXP,
				   PMQCD		arg_pMQCD,
				   PMQCHAR		arg_pcAgentBuffer,
				   long			arg_lCallType){
	long	lRc = 0;

	memset((void*)arg_pWf,0,sizeof(MXC_WORKFIELDS));

	arg_pWf->lCallType = arg_lCallType;
	arg_pWf->pMQCD = arg_pMQCD;
	arg_pWf->pcAgentBuffer = arg_pcAgentBuffer;
	arg_pWf->pMQCXP = arg_pMQCXP;

	nullDelimit(arg_pWf->szQMgr,arg_pMQCD->QMgrName,MQ_Q_MGR_NAME_LENGTH+1);
	nullDelimit(arg_pWf->szChannelName,arg_pMQCD->ChannelName,MQ_CHANNEL_NAME_LENGTH+1);
	nullDelimit(arg_pWf->szConnectionName,arg_pMQCD->ShortConnectionName,MQ_SHORT_CONN_NAME_LENGTH+1);
	nullDelimit(arg_pWf->szRemoteUserId,arg_pMQCD->RemoteUserIdentifier,MQ_USER_ID_LENGTH+1);

	arg_pWf->lOpenType = (arg_pWf->pMQCXP->ExitUserArea[MXC_EXITUSERAREA_RESPONSE_OFFSET] == MXC_REJECTED)?MXC_BROWSE: MXC_GET;

	// Set default values
//	sprintf(arg_pWf->szLogFileName,"%s.%s.%s.%s",MXC_DEFAULT_LOG_FILE_PREFIX,arg_pWf->szQMgr,arg_pWf->szChannelName,MXC_DEFAULT_LOG_FILE_SUFFIX);
//	strcpy(arg_pWf->szLogFileDirectory,MXC_DEFAULT_LOG_FILE_DIRECTORY);
	arg_pWf->sd.lMaxNumberOfInUseAttempts = MXC_DEFAULT_MAX_OBJECTS_IN_USE_ATTEMPTS;
	arg_pWf->sd.lTraceLevel = MXC_TRACE_LEVEL_NONE;
	arg_pWf->sd.lFrequencyOfPCFChecks = 1;

	// Construct correlation id for volatile data 
	arg_pWf->szVolatileCorrelId[0] = MXC_VOLATILE_PREFIX;
	arg_pWf->szVolatileCorrelId[1] = 0;
	strcat(arg_pWf->szVolatileCorrelId,arg_pWf->szChannelName);

	// Construct correlation id for bootstrap data
	arg_pWf->szBootstrapCorrelId[0] = MXC_BOOTSTRAP_PREFIX;
	arg_pWf->szBootstrapCorrelId[1] = 0;
	strcat(arg_pWf->szBootstrapCorrelId,arg_pWf->szChannelName);

	// Connect 
	lRc = connectToQMgr(arg_pWf);
	if (lRc == MXC_RC_NONE){
		// Construct the State queue name
		strcpy(arg_pWf->szDataQueueName,MXC_QUEUE_NAME_PREFIX);
		strcat(arg_pWf->szDataQueueName,arg_pWf->szChannelName);
		// Open the queue
		lRc = openQueue(arg_pWf,arg_pWf->szDataQueueName,arg_pWf->lOpenType,&(arg_pWf->dataQHobj));
	}

			

	return lRc;
}

//**********************************************************************************************************
// Connect
//
// arg_pWf			work properties
//**********************************************************************************************************
long	connectToQMgr(pMXC_WORKFIELDS arg_pWf){
	long	lRc = MXC_RC_NONE;

	MQLONG	reason;
	MQLONG	compCode;
	
	MQOD	od = {MQOD_DEFAULT};

	MQCONN(arg_pWf->szQMgr,
					&(arg_pWf->hConn),
					&compCode,
					&reason);
	if (reason == MQRC_ALREADY_CONNECTED || reason == MQRC_NONE){
	}else{
		lRc = MXC_RC_ERROR;
		writeError(arg_pWf,(char*)"Error connecting to %s. Reason=%ld",arg_pWf->szQMgr,reason);
	}

	return lRc;
}

//**********************************************************************************************************
// Open queue
//
// arg_pWf			work properties
//**********************************************************************************************************
long	openQueue(pMXC_WORKFIELDS	arg_pWf,
				  char			arg_szQName[],
				  long			arg_lOpenType,
				  MQHOBJ*		arg_pHobj){
	long	lObjectInUseAttempts=0;
	long	lRc = MXC_RC_NONE;

	MQLONG	reason;
	MQLONG	compCode;
	MQLONG	options;

	MQOD	od = {MQOD_DEFAULT};	

	// Store the name
	strncpy(od.ObjectName,arg_szQName,MQ_Q_NAME_LENGTH);

	if ((arg_pWf->sd.lTraceLevel )){
		if ((arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_DEBUG) == MXC_TRACE_LEVEL_DEBUG){
			writeLog(arg_pWf,MXC_LOG_DEBUG,(char*)"Volatile data queue is %s",arg_pWf->szDataQueueName);
		}
	}		

	// Determine open options.
	options = (arg_lOpenType == MXC_BROWSE)?MQOO_BROWSE:MQOO_INPUT_EXCLUSIVE | MQOO_OUTPUT|MQOO_BROWSE;

	// Open the queue.  . 
	// If the open is exclusive and fails due to MQRC_OBJECT_IN_USE then retry up to the lObjectInUseAttempts value
	for (lObjectInUseAttempts=0;;lObjectInUseAttempts++){
		MQOPEN(arg_pWf->hConn,
			&od,
			options,
			arg_pHobj,
			&compCode,
			&reason);
		if (reason == MQRC_NONE){
			break;
		}else{
			if (reason == MQRC_OBJECT_IN_USE){
				if (lObjectInUseAttempts >= arg_pWf->sd.lMaxNumberOfInUseAttempts){
					lRc = MXC_RC_ERROR;
					writeError(arg_pWf,(char*)"Error on MQOPEN for queue %s. MQRC_OBJECT_IN_USE has occurred for %ld attempts. Reason=%ld",arg_pWf->szDataQueueName,arg_pWf->sd.lMaxNumberOfInUseAttempts,reason);
					break;
				}
				// Sleep inbetween retries...
				waitForExclusiveLockToEnd(arg_pWf,MXC_DEFAULT_IN_USE_WAIT_MSECS);
			}else{
				lRc = MXC_RC_ERROR;
				writeError(arg_pWf,(char*)"Error on MQOPEN for queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
				break;
			}
		}
	}
	return lRc;

}

//**********************************************************************************************************
// The state queue is open exclusive by another instance of the exit.
// The chances are that the Volatile message is not on the queue (since it's being processed by the exit instance which has it open).
// So, do a Browse Wait on the queue for the Volatile message. If the other instance of the exit re-writes the Volatile message, then 
// return immediately. Otherwise, return when the Browse times out.
//
// arg_pWf				work properties
// arg_lWaitInterval	waitinterval
//**********************************************************************************************************
long	waitForExclusiveLockToEnd(pMXC_WORKFIELDS	arg_pWf,
								  long			arg_lWaitInterval){
	long	lRc = 0;

	char	szWork[10];
 
	MQLONG	reason;
	MQLONG	compCode;
	MQLONG	dataLength;

	MQGMO	gmo = {MQGMO_DEFAULT};
	MQMD	md = {MQMD_DEFAULT};

	if (arg_pWf->waitQHobj == 0){		
		// Open the queue for browse
		lRc = openQueue(arg_pWf,arg_pWf->szDataQueueName,MXC_BROWSE,&(arg_pWf->waitQHobj));
	}

	if (lRc == MXC_RC_NONE){
		// Browse volatile data
		strcpy((char*)md.CorrelId,arg_pWf->szVolatileCorrelId);
		gmo.MatchOptions = MQMO_MATCH_MSG_ID;

		gmo.WaitInterval = arg_lWaitInterval;

		gmo.Options |= MQGMO_BROWSE_FIRST|MQGMO_ACCEPT_TRUNCATED_MSG;

		// Browse. This will wait for the required number of millisecs (max) and then return
		MQGET(arg_pWf->hConn,
				arg_pWf->waitQHobj,
				&md,
				&gmo,
				sizeof(szWork),
				szWork,
				&dataLength,
				&compCode,
				&reason);
		if (reason != MQRC_NONE && reason != MQRC_NO_MSG_AVAILABLE){
			lRc = MXC_RC_ERROR;
			writeError(arg_pWf,(char*)"Error on MQGET (browse wait for InUse processing) for queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
		}
	}

	return lRc;
}

//**********************************************************************************************************
// Volatile data is data which is updated for each connection request. The assumption is that the data will be on a non-persistent message.
//
// arg_pWf			work properties
// arg_szBuffer		buffer for message data
// arg_lBufSize		buffer size
//**********************************************************************************************************
long	getVolatileData(pMXC_WORKFIELDS arg_pWf,
						char*		arg_pcBuffer,
						long*		arg_plLength){

	long	lObjectInUseAttempts=0;
	long	lRc = MXC_RC_NONE;

	MQLONG	reason;
	MQLONG	compCode;
	MQLONG	lBufferLength = *arg_plLength;
	

	MQMD	md = {MQMD_DEFAULT};
	MQGMO	gmo = {MQGMO_DEFAULT};
	MQOD	od = {MQOD_DEFAULT};

	*arg_plLength;

	// Set correlation id for volatile data
	strcpy((char*)md.CorrelId,arg_pWf->szVolatileCorrelId);

	gmo.Options = MQGMO_FAIL_IF_QUIESCING;

	if (arg_pWf->lOpenType == MXC_BROWSE){
		gmo.Options |= MQGMO_BROWSE_FIRST;
	}

	gmo.MatchOptions = MQMO_MATCH_CORREL_ID;

	// Get Volatile message (if it exists)
	MQGET(arg_pWf->hConn,
				arg_pWf->dataQHobj,
				&md,
				&gmo,
				lBufferLength,
				arg_pcBuffer,
				arg_plLength,
				&compCode,
				&reason);
	if (reason == MQRC_NONE){
		if (*arg_plLength == 0){
			lRc = MXC_RC_NO_DATA_AVAILABLE;
		}
	}else{
		if (reason == MQRC_NO_MSG_AVAILABLE){
			lRc = MXC_RC_NO_DATA_AVAILABLE;
			if ((arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_INFO) == MXC_TRACE_LEVEL_INFO){
				writeLog(arg_pWf,MXC_LOG_INFO,(char*)"No volatile data available on queue %s.  Bootstrap data will be used",arg_pWf->szDataQueueName);
			}
		}else{
			lRc = MXC_RC_ERROR;
			writeError(arg_pWf,(char*)"Error reading queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
		}
	}

	return lRc;


}

//**********************************************************************************************************
// Get bootstrap data. Bootstrap data is an initial set of values which are used to set up the exit environment (including the volatile data message)
// The bootstrap data is contained in a persistent message
//
// arg_pWf			work properties
// arg_szBuffer		buffer for message data
// arg_lBufSize		buffer size
//**********************************************************************************************************
long	getBootstrapData(pMXC_WORKFIELDS	arg_pWf,
						 char*			arg_pcBuffer,
						 long*			arg_plLength){

	long	lObjectInUseAttempts=0;
	long	lRc = MXC_RC_NONE;

	MQLONG	reason;
	MQLONG	compCode;
	MQLONG	lBufferLength = *arg_plLength;

	MQMD	md = {MQMD_DEFAULT};
	MQGMO	gmo = {MQGMO_DEFAULT};
	MQOD	od = {MQOD_DEFAULT};

	*arg_plLength;

	// Set correlation id for bootstrap data
	strcpy((char*)md.CorrelId,arg_pWf->szBootstrapCorrelId);

	gmo.Options = MQGMO_FAIL_IF_QUIESCING|MQGMO_BROWSE_FIRST;

	gmo.MatchOptions = MQMO_MATCH_CORREL_ID;

	// Get bootstrap message
	MQGET(arg_pWf->hConn,
		arg_pWf->dataQHobj,
		&md,
		&gmo,
		lBufferLength,
		arg_pcBuffer,
		arg_plLength,
		&compCode,
		&reason);
	if (reason == MQRC_NONE){
		;
	}else{
		if (reason == MQRC_NO_MSG_AVAILABLE){
			// This is an error for bootstrap
			lRc = MXC_RC_NO_DATA_AVAILABLE;
		}else{
			lRc = MXC_RC_ERROR;
		}
		writeError(arg_pWf,(char*)"Error browsing queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
	}

	return lRc;
}

//**********************************************************************************************************
// Decide whether to allow the connection
// Update the volatile data store accordingly
//
// arg_pWf							work properties
// arg_lDataGotFromBootstrapStore	true if data is from bootstrap msg; false if data is from volatile message
//**********************************************************************************************************
long	decideAndUpdate(pMXC_WORKFIELDS arg_pWf, long arg_lDataGotFromBootstrapStore){
	long	lRc = MXC_RC_NONE;
	long	lDecision=0;
	long	lEvent=0;

	// Make the decision
	lRc = connectionDecision(arg_pWf,arg_lDataGotFromBootstrapStore,&lDecision,&lEvent); 
	if ( lDecision == MXC_RC_REFUSE_CONNECTION){
		// The connection request is refused... 
		if ((arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_INFO) == MXC_TRACE_LEVEL_INFO){
			if (lRc == MXC_RC_NONE){
				writeLog(arg_pWf,MXC_LOG_INFO,(char*)"*** Connection request refused for connection name <%s> using channel %s. Existing=%ld. Max=%ld",arg_pWf->szConnectionName,arg_pWf->szChannelName,arg_pWf->sd.lExistingConnections,arg_pWf->sd.lMaxConnections);
			}else{
				writeLog(arg_pWf,MXC_LOG_INFO,(char*)"*** Connection request refused for connection name <%s> using channel %s due to previous error. Existing=%ld. Max=%ld",arg_pWf->szConnectionName,arg_pWf->szChannelName,arg_pWf->sd.lExistingConnections,arg_pWf->sd.lMaxConnections);
			}
		}

		arg_pWf->pMQCXP->ExitResponse = MQXCC_SUPPRESS_FUNCTION;

		arg_pWf->pMQCXP->ExitUserArea[MXC_EXITUSERAREA_RESPONSE_OFFSET] = MXC_REJECTED;
	}else{
		// The connection request is allowed... 
		if ((arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_INFO) == MXC_TRACE_LEVEL_INFO){
			writeLog(arg_pWf,MXC_LOG_INFO,(char*)"*** Connection request allowed for connection name <%s> using channel %s. Existing=%ld. Max=%ld",arg_pWf->szConnectionName,arg_pWf->szChannelName,arg_pWf->sd.lExistingConnections,arg_pWf->sd.lMaxConnections);
		}

		// Update the number of connected channels to include the current request
		arg_pWf->sd.lExistingConnections++;

		arg_pWf->pMQCXP->ExitResponse = MQXCC_OK;

		arg_pWf->pMQCXP->ExitUserArea[MXC_EXITUSERAREA_RESPONSE_OFFSET] = MXC_ACCEPTED;
	}

	// NOTE - only rewrite if everything is ok
	// If any errors, then don't rewrite - let next exit invocation re-initialize everything 
	// (i.e. in the absence of a volatile message it will read the Bootstrap message)
	if (lRc == MXC_RC_NONE){
		if (arg_pWf->sd.lEventFlag){
			// NOTE - mqEvent must be called after createStateDataString and eventDecision must be called before...
			lEvent = eventDecision(arg_pWf,arg_pWf->sd.lExistingConnections);
		}

		// Construct the message
		createStateDataString(arg_pWf->sd,arg_pWf->szUpdatedVolatileData,MXC_VOLATILE_PREFIX);

		if (lEvent && (arg_pWf->sd.lEventFlag & MXC_EVENTS_MQ_ACTIVE) == MXC_EVENTS_MQ_ACTIVE){
			mqEvent(arg_pWf,MXC_CONNECTION_REQUEST);
		}

		// Rewrite volatile data onto the State queue
		reWriteVolatileData(arg_pWf);
	}
	
	return lRc;
}

//**********************************************************************************************************
// Write the volatile data back to the data store queue
//
// arg_pWf							work properties
//**********************************************************************************************************
long	reWriteVolatileData(pMXC_WORKFIELDS arg_pWf){

	long	lRc = MXC_RC_NONE;

	MQMD	md = {MQMD_DEFAULT};
	MQPMO	pmo = {MQPMO_DEFAULT};

	MQLONG	compcode;
	MQLONG	reason;

	if ((arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_DEBUG) == MXC_TRACE_LEVEL_DEBUG){
		writeLog(arg_pWf,MXC_LOG_DEBUG,(char*)"Rewrite volatile data <%s>",arg_pWf->szUpdatedVolatileData);
	}

	strcpy((char*)md.CorrelId,arg_pWf->szVolatileCorrelId);

	// Use PERSISTENCE_AS_Q_DEF, to allow for the volatile data to be persisted if really desired  (not recommended...)
	md.Persistence = MQPER_PERSISTENCE_AS_Q_DEF;

	memcpy(md.Format,MQFMT_STRING,MQ_FORMAT_LENGTH);

	pmo.Options = MQGMO_FAIL_IF_QUIESCING;
	
	MQPUT(arg_pWf->hConn,
		arg_pWf->dataQHobj,
		&md,
		&pmo,
		strlen(arg_pWf->szUpdatedVolatileData),
		arg_pWf->szUpdatedVolatileData,
		&compcode,
		&reason);
	if (reason != MQRC_NONE){
		lRc = MXC_RC_ERROR;
		writeError(arg_pWf,(char*)"Error writing volatile data to queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
	}


	return lRc;
}

//**********************************************************************************************************
// Close data store queue
//
// arg_pWf							work properties
//**********************************************************************************************************
long	closeQ(pMXC_WORKFIELDS arg_pWf){

	long	lRc = MXC_RC_NONE;

	MQLONG	compcode;
	MQLONG	reason;

	if (arg_pWf->dataQHobj){
		MQCLOSE(arg_pWf->hConn,
			&(arg_pWf->dataQHobj),
			0,
			&compcode,
			&reason);
		if (reason != MQRC_NONE){
			lRc = MXC_RC_ERROR;
			writeError(arg_pWf,(char*)"Error closing data queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
		}
	}

	if (arg_pWf->waitQHobj){
		MQCLOSE(arg_pWf->hConn,
			&(arg_pWf->waitQHobj),
			0,
			&compcode,
			&reason);
		if (reason != MQRC_NONE){
			lRc = MXC_RC_ERROR;
			writeError(arg_pWf,(char*)"Error closing in-use-wait queue %s. Reason=%ld",arg_pWf->szDataQueueName,reason);
		}
	}

	return lRc;
}


//**********************************************************************************************************
// Construct and send the relevant PCF event message
//
// arg_pWf							work properties
// arg_lContext						specifies whether exit is called for connection or termination
//**********************************************************************************************************
long	mqEvent(pMXC_WORKFIELDS arg_pWf,long	arg_lContext){

	char*		pcEventBuffer = 0;
	char*		pcTemp;
	char*		pcStateData = 0;
	char		szConnectionName[MQ_CONN_NAME_LENGTH+1];
	char		szTemp[50];

	long		lRc = MXC_RC_NONE;
	long		lQMgrStrucLength;
	long		lChannelNameStrucLength;
	long		lConnectionNameStrucLength;
	long		lStateStrucLength;	
	long		lStateDataLength;
	long		lEventBufferLength = 0;


	pMXC_PCFSTRINGHEADER		pPcfHeader;
	pMXC_PCFSTRINGVARIABLEHEADER		pPcfVariable;

	MQCFH		defaultMQCFH = {MQCFH_DEFAULT};
	MQCFST		defaultMQCFST = {MQCFST_DEFAULT};
	MQCFIN		defaultMQCFIN = {MQCFIN_DEFAULT};

	nullDelimit(szConnectionName,arg_pWf->pMQCD->ConnectionName,MQ_CHANNEL_NAME_LENGTH+1);

	pcStateData = (char*)calloc(lStateDataLength = sizeof(MXC_PCFSTRINGHEADER) + sizeof(MXC_PCFSTRINGVARIABLEHEADER) + strlen(arg_pWf->szUpdatedVolatileData) ,1);
	if (!pcStateData){
		writeError(arg_pWf,(char*)"Error with calloc for %ld bytes of storage for statedata (for event message).",lStateDataLength );
		lRc = MXC_RC_ERROR;
	}else{
		// Calculate struclengths for each event parameter block (must be a multiple of 4)
		lStateStrucLength = MQCFST_STRUC_LENGTH_FIXED + lStateDataLength + 4-(lStateDataLength%4?lStateDataLength%4:4);
		lQMgrStrucLength = MQCFST_STRUC_LENGTH_FIXED + strlen(arg_pWf->szQMgr) + 4-(strlen(arg_pWf->szQMgr)%4?strlen(arg_pWf->szQMgr)%4:4); 
		lChannelNameStrucLength = MQCFST_STRUC_LENGTH_FIXED + strlen(arg_pWf->szChannelName) + 4-(strlen(arg_pWf->szChannelName)%4?strlen(arg_pWf->szChannelName)%4:4);
		lConnectionNameStrucLength = MQCFST_STRUC_LENGTH_FIXED + strlen(szConnectionName) + 4-(strlen(szConnectionName)%4?strlen(szConnectionName)%4:4);
	
		pcEventBuffer  = (char*)calloc(lEventBufferLength = MQCFH_STRUC_LENGTH + 
												lStateStrucLength +
												lQMgrStrucLength +
												lChannelNameStrucLength +
												lConnectionNameStrucLength,1);
		if (!pcEventBuffer){
			writeError(arg_pWf,(char*)"Error with calloc for %ld bytes of storage for event message.",lEventBufferLength);
			lRc = MXC_RC_ERROR;
		}
	}

	if (lRc == MXC_RC_NONE){

		pPcfHeader = (pMXC_PCFSTRINGHEADER)pcStateData;

		if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_OK){
			memcpy(pPcfHeader->eventType,MXC_EVENT_STATE_OK_TEXT,sizeof(pPcfHeader->eventType));
		}else{
			if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_WARN){
				memcpy(pPcfHeader->eventType,MXC_EVENT_STATE_WARN_TEXT,sizeof(pPcfHeader->eventType));
			}else{
				if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_MAX){
					memcpy(pPcfHeader->eventType,MXC_EVENT_STATE_MAX_TEXT,sizeof(pPcfHeader->eventType));
				}else{
					writeError(arg_pWf,(char*)"Unknown event state value of %ld",arg_pWf->sd.lEventState);
					lRc =  MXC_RC_ERROR;
				}
			}
		}
	}

	if (lRc == MXC_RC_NONE){
	
		if (arg_lContext == MXC_CONNECTION_REQUEST){
			memcpy(pPcfHeader->eventContext,MXC_CONNECTION_REQUEST_TEXT,sizeof(pPcfHeader->eventContext));
		}else{
			memcpy(pPcfHeader->eventContext,MXC_CONNECTION_TERMINATION_TEXT,sizeof(pPcfHeader->eventContext));
		}

		pPcfVariable = (pMXC_PCFSTRINGVARIABLEHEADER)(pcStateData + sizeof(MXC_PCFSTRINGHEADER));

		sprintf(szTemp,"%ld",strlen(arg_pWf->szUpdatedVolatileData));
		memcpy(pPcfVariable->varLength,MXC_PCFSTRINGVARIABLEHEADER_LENGTH_DEFAULT,sizeof(pPcfVariable->varLength));
		memcpy(&(pPcfVariable->varLength[sizeof(pPcfVariable->varLength) - strlen(szTemp)]),szTemp,strlen(szTemp));

		pcTemp = (char*)pPcfVariable + sizeof(MXC_PCFSTRINGVARIABLEHEADER);
		memcpy(pcTemp,arg_pWf->szUpdatedVolatileData,strlen(arg_pWf->szUpdatedVolatileData));

		// Reset pcTemp to pcEventBuffer 
		pcTemp = pcEventBuffer ;
		// Construct MQCFH header
		// NOTE - the MQCFH.Reason codes are user defined in maxconn.h (MXC_PCF_REASON_CONNECTIONS_OK etc)
		memcpy(pcTemp,(char*)&defaultMQCFH,MQCFH_STRUC_LENGTH);

		((MQCFH*)pcTemp)->ParameterCount =	4;
		((MQCFH*)pcTemp)->Command =			MQCMD_CHANNEL_EVENT;
		if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_OK){
			((MQCFH*)pcTemp)->CompCode = MQCC_OK;
			((MQCFH*)pcTemp)->Reason = MXC_PCF_REASON_CONNECTIONS_OK;

		}else{
			if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_WARN){
				((MQCFH*)pcTemp)->CompCode = MQCC_WARNING;
				((MQCFH*)pcTemp)->Reason = MXC_PCF_REASON_WARN_THRESHOLD_REACHED;
			}else{
				((MQCFH*)pcTemp)->CompCode = MQCC_FAILED;
				((MQCFH*)pcTemp)->Reason = MXC_PCF_REASON_MAX_CONNECTIONS_REACHED;
			}
		}

		pcTemp += MQCFH_STRUC_LENGTH;

		// Construct MQCFST block for state info
		memcpy(pcTemp,(char*)&defaultMQCFST,MQCFST_STRUC_LENGTH_FIXED);
		((MQCFST*)pcTemp)->Parameter = MQCACF_STRING_DATA;
		((MQCFST*)pcTemp)->StrucLength = lStateStrucLength;
		((MQCFST*)pcTemp)->StringLength = lStateDataLength;
		memcpy((char*)((MQCFST*)pcTemp)->String,pcStateData,lStateDataLength);

		pcTemp += lStateStrucLength; 

		// Construct MQCFST block for qmgr name
		memcpy(pcTemp,(char*)&defaultMQCFST,MQCFST_STRUC_LENGTH_FIXED);
		((MQCFST*)pcTemp)->Parameter = MQCA_Q_MGR_NAME;
		((MQCFST*)pcTemp)->StrucLength = lQMgrStrucLength;
		((MQCFST*)pcTemp)->StringLength = strlen(arg_pWf->szQMgr);
		memcpy((char*)((MQCFST*)pcTemp)->String,arg_pWf->szQMgr,strlen(arg_pWf->szQMgr));

		pcTemp += lQMgrStrucLength; 

		// Construct MQCFST block for channel name
		memcpy(pcTemp,(char*)&defaultMQCFST,MQCFST_STRUC_LENGTH_FIXED);
		((MQCFST*)pcTemp)->Parameter = MQCACH_CHANNEL_NAME;
		((MQCFST*)pcTemp)->StrucLength = lChannelNameStrucLength;
		((MQCFST*)pcTemp)->StringLength = strlen(arg_pWf->szChannelName);
		memcpy((char*)((MQCFST*)pcTemp)->String,arg_pWf->szChannelName,strlen(arg_pWf->szChannelName));

		pcTemp += lChannelNameStrucLength;

		// Construct MQCFST block for connection name
		memcpy(pcTemp,(char*)&defaultMQCFST,MQCFST_STRUC_LENGTH_FIXED);
		((MQCFST*)pcTemp)->Parameter = MQCACH_CONNECTION_NAME;
		((MQCFST*)pcTemp)->StrucLength = lConnectionNameStrucLength;
		((MQCFST*)pcTemp)->StringLength = strlen(szConnectionName);
		memcpy((char*)((MQCFST*)pcTemp)->String,szConnectionName,strlen(szConnectionName));

		// Send the event message
		lRc = sendMqEventMessage(arg_pWf,pcEventBuffer,lEventBufferLength);
	}
	if (pcStateData){
		free(pcStateData);
	}
	if (pcEventBuffer){
		free(pcEventBuffer);
	}

	return 0;
}

//**********************************************************************************************************
// Send the mq event message
//
// arg_pWf							work properties
// arg_szBuffer						message buffer
// arg_lBufferLength				buffer length
//**********************************************************************************************************
long	sendMqEventMessage(pMXC_WORKFIELDS	arg_pWf,char*	arg_pcBuffer,long	arg_lBufferLength){
	long	lRc = MXC_RC_NONE;

	MQMD	md = {MQMD_DEFAULT};
	MQOD	od = {MQOD_DEFAULT};
	MQPMO	pmo = {MQPMO_DEFAULT};

	MQLONG	compcode;
	MQLONG	reason;

	strncpy(od.ObjectName,arg_pWf->sd.szEventQueue,MQ_Q_NAME_LENGTH);

	// Note that the correlid for event messages is channel name. 
	// This might be useful (but only if the event is read under a similar CCSID to the local CCSID)
	strcpy((char*)md.CorrelId,arg_pWf->szChannelName);

	memcpy(md.Format,MQFMT_PCF,MQ_FORMAT_LENGTH);

	// Event messages are persistent
	md.Persistence = MQPER_PERSISTENT;

	pmo.Options = MQPMO_FAIL_IF_QUIESCING ;
	
	MQPUT1(arg_pWf->hConn,
		&od,
		&md,
		&pmo,
		arg_lBufferLength,
		arg_pcBuffer,
		&compcode,
		&reason);
	if (reason != MQRC_NONE){
		lRc = MXC_RC_ERROR;
		writeError(arg_pWf,(char*)"Error on MQPUT1 for queue %s. Reason=%ld",arg_pWf->sd.szEventQueue,reason);
	}

	return lRc;
}


//**********************************************************************************************************
// Parse the data retrieved from the data store (volatile or bootstrap)
//
// arg_pWf							work properties
// arg_szBuffer						state data 
//**********************************************************************************************************
long	extractValues(pMXC_WORKFIELDS arg_pWf,char*	arg_pcBuffer,long	arg_lBufferLength){
	char	szError[500];

	long lRc = MXC_RC_NONE;


	lRc = extractFromStateDataString(&(arg_pWf->sd),arg_pcBuffer,arg_lBufferLength,szError);
	
	if (lRc != MXC_RC_NONE){
		writeError(arg_pWf,szError);
	}

	strcpy(arg_pWf->szLogFile,arg_pWf->sd.szLogFileDirectory);
	strcat(arg_pWf->szLogFile,MXC_FILE_DELIMITER);
	strcat(arg_pWf->szLogFile,arg_pWf->sd.szLogFileName);

	return lRc;
}




//**********************************************************************************************************
// decide whether to allow the existing connection request
//
// arg_pWf							work properties
// arg_lDataGotFromBootstrapStore	>0 if data got from bootstrap msg; 0 if data got from volatile msg
// arg_plDecision					specifies whether to accept or refuse connection request
// arg_plEvent						not used (always set to 0 in current version of maxconn)
//**********************************************************************************************************
long	connectionDecision(pMXC_WORKFIELDS	arg_pWf,
						   long			arg_lDataGotFromBootstrapStore,
						   long*		arg_plDecision,
						   long*		arg_plEvent){
	long	lRc = MXC_RC_NONE;
	long	lConnectionsTemp = arg_pWf->sd.lExistingConnections;

	*arg_plEvent = 0;

	// If the existing channel count (minus the current request) equals or is more than max channels then confirm the existing count before rejecting.
	// NOTE - additionally, only confirm existing count if the PCF frequency check parameter (part of the State data) allows it 
	if (arg_pWf->sd.lExistingConnections >= arg_pWf->sd.lMaxConnections ){
		if (arg_pWf->sd.lFrequencyOfPCFChecks && ++(arg_pWf->sd.lTimesAtConnMaxSinceLastPCFCheck) >= arg_pWf->sd.lFrequencyOfPCFChecks){
			lRc = getActualCurrentNumberOfRunningChannels(arg_pWf,&lConnectionsTemp);
			//current running chstatuses includes the current MCA instance, so subtract that to get existing connections
			if (lConnectionsTemp){
				lConnectionsTemp--;
			}
			if (lRc == MXC_RC_NONE && (arg_pWf->sd.lTraceLevel & MXC_TRACE_LEVEL_DEBUG) == MXC_TRACE_LEVEL_DEBUG){
				writeLog(arg_pWf,MXC_LOG_DEBUG,(char*)"TimesAtConnMaxSinceLastPCFCheck=%ld. FrequencyOfPCFChecks=%ld.  PCF check was made",arg_pWf->sd.lTimesAtConnMaxSinceLastPCFCheck,arg_pWf->sd.lFrequencyOfPCFChecks);
				if (lConnectionsTemp != arg_pWf->sd.lExistingConnections){
					writeLog(arg_pWf,MXC_LOG_DEBUG,(char*)"Volatile data indicates that ExistingConnections=MaxConnections (%ld). Number of MQCHS_RUNNING channels (%ld) differed from volatile ExistingConnections value, so ExistingConnections value was updated.",arg_pWf->sd.lMaxConnections,lConnectionsTemp);
				}
			}
			if (arg_pWf->sd.lTimesAtConnMaxSinceLastPCFCheck){
				arg_pWf->sd.lTimesAtConnMaxSinceLastPCFCheck = 0;
			}
		}else{
			arg_pWf->sd.lTimesAtConnMaxSinceLastPCFCheck = 0;
		}

	}else{
		// If the state data was taken from the bootstrap store then we have no idea of the current environment so get the actual number of channels.
		if ( arg_lDataGotFromBootstrapStore){
			lRc = getActualCurrentNumberOfRunningChannels(arg_pWf,&lConnectionsTemp);
			//current running chstatuses includes the current MCA instance, so subtract that to get existing connections
			if (lConnectionsTemp){
				lConnectionsTemp--;
			}
		}
	}

	if (lRc == MXC_RC_NONE){
		arg_pWf->sd.lExistingConnections = lConnectionsTemp;
		// If the existing channels is equal to or exceeds max channel level, then refuse the current connection request
		*arg_plDecision = (arg_pWf->sd.lExistingConnections >= arg_pWf->sd.lMaxConnections)?MXC_RC_REFUSE_CONNECTION:MXC_RC_ALLOW_CONNECTION;
	}else{	
		*arg_plDecision = MXC_RC_REFUSE_CONNECTION;
	}
	
	return lRc;

}

//**********************************************************************************************************
// Decide whether an event is due
//
// arg_pWf				work properties
// arg_lConnections		the number of current connections including the current request
//**********************************************************************************************************
long	eventDecision(pMXC_WORKFIELDS arg_pWf,long arg_lConnections){
	long	lGeneratedEvent = 0;

	// Do the number of connects exceed or equal the max value?
	if (arg_lConnections >= arg_pWf->sd.lMaxConnections){
		if (arg_pWf->sd.lEventState != MXC_EVENT_STATE_MAX){
			lGeneratedEvent = MXC_EVENT_STATE_MAX;
		} 
	}else{
		// Do the number of connects exceed or equal the warn value (if a warn value exists)?
		if (arg_pWf->sd.lWarnThreshold && arg_lConnections < arg_pWf->sd.lMaxConnections && arg_lConnections  >= arg_pWf->sd.lWarnThreshold){
			if (arg_pWf->sd.lEventState != MXC_EVENT_STATE_WARN){
				lGeneratedEvent = MXC_EVENT_STATE_WARN;
			}
		}else{
			if (arg_pWf->sd.lEventState != MXC_EVENT_STATE_OK ){
				lGeneratedEvent = MXC_EVENT_STATE_OK;
			}
		}
	}

	if (lGeneratedEvent ){
		arg_pWf->sd.lEventState = lGeneratedEvent;
		if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_OK){
			writeLog(arg_pWf,MXC_LOG_EVENT,(char*)"%s. Number of connections (%ld) is OK",MXC_EVENT_STATE_OK_TEXT,arg_pWf->sd.lExistingConnections );
		}else{
			if (arg_pWf->sd.lEventState == MXC_EVENT_STATE_WARN){
				writeLog(arg_pWf,MXC_LOG_EVENT,(char*)"%s. Number of connections (%ld ) equals or exceeds warn threshold (%ld)",MXC_EVENT_STATE_WARN_TEXT,arg_pWf->sd.lExistingConnections,arg_pWf->sd.lWarnThreshold);
			}else{
				writeLog(arg_pWf,MXC_LOG_EVENT,(char*)"%s. Number of connections (%ld) equals or exceeds max (%ld)",MXC_EVENT_STATE_MAX_TEXT,arg_pWf->sd.lExistingConnections,arg_pWf->sd.lMaxConnections);
			}
		}
	}
	return lGeneratedEvent;
}

//**********************************************************************************************************
// Use the admin interface to query the channel status of instances of the channel in question, and count the number with MQCHS_RUNNING
//
// arg_pWf							work properties
// arg_plNumberOfChannels			number of extant channels with the relevant SvrConn name
//**********************************************************************************************************
long	getActualCurrentNumberOfRunningChannels(pMXC_WORKFIELDS arg_pWf,long* arg_plNumberOfChannels){

	MQLONG	reasonCode;
	MQLONG	completionCode;

	long	lRc=MXC_RC_NONE;

	MQHBAG adminBag = MQHB_UNUSABLE_HBAG;
	MQHBAG responseBag = MQHB_UNUSABLE_HBAG;
	MQHBAG qAttrsBag = MQHB_UNUSABLE_HBAG;

	*arg_plNumberOfChannels = 0;

	do{
		lRc = bagSetup(arg_pWf,&adminBag,&responseBag,&qAttrsBag);
		if (lRc != MXC_RC_NONE){
			break;
		}
		
		mqAddString(adminBag, MQCACH_CHANNEL_NAME, MQBL_NULL_TERMINATED, arg_pWf->szChannelName, &completionCode, &reasonCode);
		if (reasonCode != MQRC_NONE){
			writeError(arg_pWf,(char*)"Error with mqAddString for MQCACH_CHANNEL_NAME of %s. Reason=%ld",arg_pWf->szChannelName,reasonCode);
			lRc = MXC_RC_ERROR;	
			break;
		}

		mqAddInquiry(adminBag, MQIACH_CHANNEL_STATUS, &completionCode, &reasonCode);
		if (reasonCode != MQRC_NONE){
			writeError(arg_pWf,(char*)"Error with mqAddInquiry for MQIACH_CHANNEL_STATUS. Reason=%ld",reasonCode);
			lRc = MXC_RC_ERROR;	
			break;
		}

		mqExecute(arg_pWf->hConn,
             MQCMD_INQUIRE_CHANNEL_STATUS,
             MQHB_NONE,
             adminBag,
             responseBag,
             MQHO_NONE,
             MQHO_NONE,
             &completionCode,
             &reasonCode);
 
		if (reasonCode != MQRC_NONE){
			if (reasonCode == MQRCCF_COMMAND_FAILED){
				MQLONG mqExecuteCC;
				MQLONG mqExecuteRC;
				MQHBAG errorBag;

				mqInquireBag(responseBag, MQHA_BAG_HANDLE, 0, &errorBag, &completionCode, &reasonCode);
				if (reasonCode != MQRC_NONE){
					writeError(arg_pWf,(char*)"Error with mqInquireBag . Reason=%ld",reasonCode);
					lRc = MXC_RC_ERROR;	
					break;
				}
				mqInquireInteger(errorBag,
							MQIASY_COMP_CODE,
							MQIND_NONE,
							&mqExecuteCC,
							&completionCode,
							&reasonCode );
				if (reasonCode != MQRC_NONE){
					writeError(arg_pWf,(char*)"Error with mqInquireInteger (CC). Reason=%ld",reasonCode);
					lRc = MXC_RC_ERROR;	
					break;
				}

				mqInquireInteger(errorBag,
							MQIASY_REASON,
							MQIND_NONE,
							&mqExecuteRC,
							&completionCode,
							&reasonCode);
				if (reasonCode != MQRC_NONE){
					writeError(arg_pWf,(char*)"Error with mqInquireInteger (RC). Reason=%ld",reasonCode);
					lRc = MXC_RC_ERROR;	
					break;
				}

				if (mqExecuteRC == MQRCCF_CHL_STATUS_NOT_FOUND){
					*arg_plNumberOfChannels = 0;
				}else{
					writeError(arg_pWf,(char*)"Error with mqExecute. ExecuteReason=%ld",mqExecuteRC);
					lRc = MXC_RC_ERROR;	
					break;
				}
			}else{
				writeError(arg_pWf,(char*)"Error with mqExecute. Reason=%ld",reasonCode);
				lRc = MXC_RC_ERROR;	
				break;
			}
		}

		if (lRc == MXC_RC_NONE){
			long	lNumberOfResponseBags;
			mqCountItems(responseBag, MQHA_BAG_HANDLE, &lNumberOfResponseBags, &completionCode, &reasonCode);
			if (reasonCode != MQRC_NONE){
				writeError(arg_pWf,(char*)"Error with mqCountItems. Reason=%ld",reasonCode);
				lRc = MXC_RC_ERROR;	
				break;
			}
			if (lNumberOfResponseBags){
				long	l1;
				long	lItemCount;

				// one bag per channel instance - loop thru them
				for (l1=0;l1< lNumberOfResponseBags;l1++){
					mqInquireBag(responseBag, MQHA_BAG_HANDLE, l1, &qAttrsBag, &completionCode, &reasonCode);
					if (reasonCode != MQRC_NONE){
						writeError(arg_pWf,(char*)"Error with mqInquireBag. Reason=%ld",reasonCode);
						lRc = MXC_RC_ERROR;	
						break;
					}
					mqCountItems(qAttrsBag,MQSEL_ALL_SELECTORS ,&lItemCount,&completionCode, &reasonCode);
					if (reasonCode != MQRC_NONE){
						writeError(arg_pWf,(char*)"Error with mqCountItems . Reason=%ld",reasonCode);
						lRc = MXC_RC_ERROR;	
						break;
					}
					if (lItemCount){
						long	lStatus;
						mqInquireInteger(qAttrsBag, MQIACH_CHANNEL_STATUS, MQIND_NONE,&lStatus,
								&completionCode, &reasonCode);
						if (reasonCode != MQRC_NONE){
							writeError(arg_pWf,(char*)"Error with mqInquireInteger MQIACH_CHANNEL_STATUS. Reason=%ld",reasonCode);
							lRc = MXC_RC_ERROR;	
							break;
						}
	
						// If the status is RUNNING, then increment counter
						if (lStatus == MQCHS_RUNNING){
							(*arg_plNumberOfChannels)++; 
						}
					}
				}
			}
		}
	}while(0);

	return lRc;
}

//**********************************************************************************************************
// setup admin interface objects
//**********************************************************************************************************
long	bagSetup(pMXC_WORKFIELDS arg_pWf,MQHBAG*const arg_pAdminBag,MQHBAG*const arg_pResponseBag,MQHBAG*const arg_pQAttrsBag){

	long	lRc = MXC_RC_NONE;

	MQLONG reasonCode;
	MQLONG completionCode;

	do{
		if (*arg_pAdminBag != MQHB_UNUSABLE_HBAG){
			mqClearBag(*arg_pAdminBag, &completionCode, &reasonCode);
			if (reasonCode != MQRC_NONE){
				writeError(arg_pWf,(char*)"Error with mqClearBag . Reason=%ld",reasonCode);
				lRc = MXC_RC_ERROR;
				break;
			}
		}
		mqCreateBag(MQCBO_ADMIN_BAG, arg_pAdminBag, &completionCode, &reasonCode);
		if (reasonCode != MQRC_NONE){
			writeError(arg_pWf,(char*)"Error with mqCreateBag . Reason=%ld",reasonCode);
			lRc = MXC_RC_ERROR;
			break;
		}
	
		if (*arg_pResponseBag != MQHB_UNUSABLE_HBAG){
			mqClearBag(*arg_pResponseBag, &completionCode, &reasonCode);
			if (reasonCode != MQRC_NONE){
				writeError(arg_pWf,(char*)"Error with mqClearBag . Reason=%ld",reasonCode);
				lRc = MXC_RC_ERROR;
				break;
			}
		}
		mqCreateBag(MQCBO_ADMIN_BAG, arg_pResponseBag, &completionCode, &reasonCode);
		if (reasonCode != MQRC_NONE){
			writeError(arg_pWf,(char*)"Error with mqCreateBag . Reason=%ld",reasonCode);
			lRc = MXC_RC_ERROR;
			break;
		}

		if (*arg_pQAttrsBag != MQHB_UNUSABLE_HBAG){
			mqClearBag(*arg_pQAttrsBag, &completionCode, &reasonCode);
			if (reasonCode != MQRC_NONE){
				writeError(arg_pWf,(char*)"Error with mqClearBag . Reason=%ld",reasonCode);
				lRc = MXC_RC_ERROR;
				break;
			}
			*arg_pQAttrsBag = MQHB_UNUSABLE_HBAG;
		}
	}while(0);

	return lRc;

}



//**********************************************************************************************************
// Interrogate a fixed size array. Replace the first space with a null delimiter.  If no space exists, set the byte following the array to null. 
//**********************************************************************************************************
void	nullDelimit(char arg_szString[],char arg_acArray[],long arg_lMaxBufferSize ){
	char* pcTemp;

	memcpy(arg_szString,arg_acArray,arg_lMaxBufferSize-1);
	pcTemp = strstr(arg_szString," ");
	if (pcTemp){
		*pcTemp=0;
	}else{
		arg_szString[arg_lMaxBufferSize-1] = 0;
	}
}


//**********************************************************************************************************
// Write a trace message
//**********************************************************************************************************
void	writeLog(pMXC_WORKFIELDS arg_pWf,char arg_szLogType[],char arg_szData[],...){

	char	szWork1[5000];
	char	szWork2[5000];
	char	szTime[100];

	time_t	tp;
	struct tm*	now;
	va_list va;

	va_start(va,arg_szData);
	vsprintf(szWork1,arg_szData,va);
	va_end(va);

	tp = time(0);
	now = localtime(&tp);
	strftime(szTime,sizeof(szTime)-1,"%c",now);

	sprintf(szWork2,"%s  %s. %s. %s.  QMgr=%s. Channel=%s. RemoteUserId=%s.  ConnName=%s",szTime,arg_pWf->lCallType == MXC_REQUEST?MXC_CONNECTION_REQUEST_TEXT:MXC_CONNECTION_TERMINATION_TEXT,arg_szLogType,szWork1,arg_pWf->szQMgr,arg_pWf->szChannelName,arg_pWf->szRemoteUserId,arg_pWf->szConnectionName);
	writeToFile(arg_pWf,szWork2);

}

//**********************************************************************************************************
// Write an error message
//**********************************************************************************************************
void	writeError(pMXC_WORKFIELDS arg_pWf,char arg_szData[],...){

	char	szWork1[5000];
	char	szWork2[5000];
	char	szTime[100];
	time_t	tp;
	struct tm*	now;
	va_list va;

	va_start(va,arg_szData);
	vsprintf(szWork1,arg_szData,va);
	va_end(va);

	tp = time(0);
	now = localtime(&tp);
	strftime(szTime,sizeof(szTime)-1,"%c",now);

	sprintf(szWork2,"%s. %s. %s. %s.  QMgr=%s. Channel=%s. RemoteUserId=%s  ConnName=%s.  %s",szTime,arg_pWf->lCallType == MXC_REQUEST?MXC_CONNECTION_REQUEST_TEXT:MXC_CONNECTION_TERMINATION_TEXT,MXC_LOG_ERROR,szWork1,arg_pWf->szQMgr,arg_pWf->szChannelName,arg_pWf->szRemoteUserId,arg_pWf->szConnectionName,MXC_EXIT_PROG_IDENT);
	writeToFile(arg_pWf,szWork2);

}


//**********************************************************************************************************
// Write to a file
//**********************************************************************************************************
void	writeToFile(pMXC_WORKFIELDS arg_pWf,char arg_szData[]){
	FILE*	file = 0;
	
	file = fopen(arg_pWf->szLogFile,"a+");
	if (file != 0){
		fputs(arg_szData,file);
		fputs("\n",file);
		fclose(file);
	}else{
		// Assume that the error was due to a directory problem and try writing to the root
		char	szWork[257];
		sprintf(szWork,"%s%s",MXC_ROOT_LOCATION,arg_pWf->sd.szLogFileName);
		file = fopen(szWork,"a+");
		if (file != 0){
			fputs(arg_szData,file);
			fputs("\n",file);
			fclose(file);
		}
	}
}

