//***********************************************************************************************************
//
// An Exception class
//
//
// author	Tony Madden	(tm@mqware.com)
//***********************************************************************************************************

#include	<x_exception.hpp>
#include	<stdio.h>


X_Exception::X_Exception(){
	initVariables();
	return;
	
}

//*************************************************************************************
// Construct using another X_Exception object
//*************************************************************************************
X_Exception::X_Exception(const X_Exception& arg_X){
	initVariables();

	X_Exception * pX = (X_Exception*)&arg_X;

	setup(pX->getCode(),
		pX->getSubCode(),
		pX->getText(),
		pX->getThrownBy(),
		pX->getInnerException());
 
}

//*************************************************************************************
// Construct using a string
//*************************************************************************************
X_Exception::X_Exception(string arg_s){
	initVariables();

	setup(0,0,arg_s,"",NULL);
}

//*************************************************************************************
// Construct using individual properties
//
// arg_lCode			main code
// arg_lSubCode			sub code
// arg_sText			text
// arg_sThrownBy		the throwing method
//*************************************************************************************
X_Exception::X_Exception(long arg_lCode,long arg_lSubCode,string arg_sText,string	arg_sThrownBy){
	setup(arg_lCode,arg_lSubCode,arg_sText,arg_sThrownBy,NULL);

}
//*************************************************************************************
// Construct using individual properties
//
// arg_lCode			main code
// arg_lSubCode			sub code
// arg_sText			text
///*************************************************************************************
X_Exception::X_Exception(long arg_lCode,long arg_lSubCode,string arg_sText){
	setup(arg_lCode,arg_lSubCode,arg_sText,"",NULL);

}
//*************************************************************************************
// Construct using individual properties and an X_Exception object
//
// arg_lCode			main code
// arg_lSubCode			sub code
// arg_sText			text
// arg_sThrownBy		the throwing method
// arg_X				an X_Exception object
//*************************************************************************************
X_Exception::X_Exception(long arg_lCode,long arg_lSubCode,string arg_sText,string arg_sThrownBy,const X_Exception& arg_X){
	setup(arg_lCode,arg_lSubCode,arg_sText,arg_sThrownBy,&arg_X);
}

//*************************************************************************************
// Initialization
//
// arg_lCode			main code
// arg_lSubCode			sub code
// arg_sText			text
// arg_X				an X_Exception object
//*************************************************************************************
void X_Exception::setup(long arg_lCode,long arg_lSubCode,string arg_sText,string arg_sThrownBy,const X_Exception * arg_pX){
	initVariables();
	lCode = arg_lCode;
	lSubCode = arg_lSubCode;
	sText = arg_sText;
	sThrownBy = arg_sThrownBy;
	if (arg_pX){
		setInnerException(*arg_pX);
	}
}

//*************************************************************************************
// Initialization
//
// arg_X				an X_Exception object
//*************************************************************************************
void X_Exception::setup(const X_Exception& arg_X){
	initVariables();
	lCode = ((X_Exception)arg_X).getCode();
	lSubCode = ((X_Exception)arg_X).getSubCode();
	sText = ((X_Exception)arg_X).getText();
	sThrownBy = ((X_Exception)arg_X).getThrownBy();
	if (((X_Exception)arg_X).getInnerException()){
		setInnerException(*(((X_Exception)arg_X).getInnerException()));
	}
}

//*************************************************************************************
//  Equals operator
//*************************************************************************************
X_Exception& X_Exception::operator=(const X_Exception& arg_X){
	if (this != &arg_X){
		setup((X_Exception&)arg_X);
	}
	return *this;
}


//*************************************************************************************
// Convert object properties into a string 
//*************************************************************************************
string	X_Exception::toString(){
	ostringstream	ost;

	if (sThrownBy.size()){
		ost << sThrownBy << " ";
	}
	ost << "Code=" << lCode << "  SubCode=" << lSubCode << "  Text=" << sText;

	if (pInnerException){
		ost << "  InnerException{" << pInnerException->toString() << "}";
	}

	return ost.str();
}

//*************************************************************************************
//*************************************************************************************
void	X_Exception::initVariables(){
	lCode = lSubCode = 0;
	sText="";
	sThrownBy="";
	pInnerException = NULL;

}

//*************************************************************************************
//*************************************************************************************
long X_Exception::getCode() {
	return lCode;
}

//*************************************************************************************
//*************************************************************************************
long X_Exception::getSubCode() {
	return lSubCode;
}
//*************************************************************************************
//*************************************************************************************
string X_Exception::getText() {
	return sText;
}

//*************************************************************************************
// Get the method which originally threw (or created) this exception (if set)
//*************************************************************************************
string X_Exception::getThrownBy() {
	return sThrownBy;
}

//*************************************************************************************
//*************************************************************************************
X_Exception*const X_Exception::getInnerException() {
	return ( X_Exception*const)pInnerException;
}

//*************************************************************************************
//*************************************************************************************
void X_Exception::setCode(long l){
	lCode = l;
}

//*************************************************************************************
//*************************************************************************************
void X_Exception::setSubCode(long l){
	 lSubCode = l;
}
//*************************************************************************************
//*************************************************************************************
void X_Exception::setText(string arg_s){
	 sText = arg_s;
}
//*************************************************************************************
// Set the method which created this exception object
//*************************************************************************************
void X_Exception::setThrownBy(string arg_s){
	 sThrownBy = arg_s;
}

//*************************************************************************************
//  Set an inner exception  (if one already exists then delete it). 
//*************************************************************************************
void	X_Exception::setInnerException(const X_Exception& arg_X){
	if (pInnerException){
		delete pInnerException;
		pInnerException = NULL;
	}
	pInnerException = new X_Exception(arg_X);
}

string	X_Exception::getVersion(){
	return X_EXCEPTION_VERSION;
}


